from .generate_error_dict import generate_error_dict
# from .preprocess import process_csv_file
from .single_vowel_constants import SPELL_DICT, SPELL_DICT_LENGTH, STANDARD_WORDS
from .match_word import match_word, match_sentence


import logging
import multiprocessing as mp
import os
from os.path import dirname, join

from fuzzywuzzy import process, fuzz

from .constants import ERROR_DICT, SINGLE_VOWELS, STANDARD_ADDRESSES, STANDARD_SINGLE_NAMES, FULL_ADDRRESS_ERROR_DICT
from .functions.acronym_layer import acronym_execute
from .functions.crf_layer import crf_execute
from .functions.crf_layer.constants import FULL_TAG_LIST
from .functions.crf_layer.crf_execute import list2sentence, extract_relative_tag
from .functions.spell_correction_layer import match_word, match_sentence
from .functions.utils import sentence2list, list_subtract

COMMA_HANDLE = False
BASE_DIR = dirname(dirname(__file__))
STANDARDIZER_MODEL_PATH = 'vnaddress/models'


class VNAddressStandardizer(object):
    def __init__(self, raw_address, comma_handle):
        self.raw_address = raw_address
        self.comma_handle = comma_handle

    def combined_processing(self, raw_address, comma_handle=COMMA_HANDLE):
        error_dict = ERROR_DICT
        standard_addresses = STANDARD_ADDRESSES
        full_address_error_dict = FULL_ADDRRESS_ERROR_DICT
        standard_single_names = STANDARD_SINGLE_NAMES

        # logger.info(raw_address)
        acronym_fixed_address = acronym_execute(raw_address)
        # logger.info(acronym_fixed_address)

        if comma_handle:
            address2list = sentence2list(acronym_fixed_address)
            pool = mp.Pool(mp.cpu_count())
            corrected_address2list = pool.starmap(match_word,
                                                  [(error_dict, name, standard_single_names) for name in address2list])
            pool.close()

        else:
            corrected_address = match_sentence(full_address_error_dict, acronym_fixed_address, standard_addresses)
            corrected_address2list = sentence2list(corrected_address)
            # logger.info(corrected_address2list)

        crf_result = crf_execute("{}/{}/finalized_model.sav".format(BASE_DIR, STANDARDIZER_MODEL_PATH), corrected_address2list)

        tag_list = [extract_relative_tag(item[1]) for item in crf_result]
        missing_tag_list = list_subtract(FULL_TAG_LIST, tag_list)
        # logger.info(missing_tag_list)

        name_only_crf_result = ", ".join([item[0] for item in crf_result])
        match_check = process.extractOne(name_only_crf_result, standard_addresses, scorer=fuzz.ratio)

        return crf_result, match_check, missing_tag_list

    def execute(self):
        try:
            raw_address = self.raw_address
            comma_handle = self.comma_handle if self.comma_handle else COMMA_HANDLE
            result, match_check, missing_tag_list = self.combined_processing(raw_address, comma_handle)

            if match_check[1] < 100:
                result, match_check, _ = self.combined_processing(match_check[0], comma_handle)
                if "PX" in missing_tag_list:
                    result = result[1:]
                    if "QH" in missing_tag_list:
                        result = result[1:]

            result_data = {
                "result": list2sentence(result),
                "match": {
                    "match_address": match_check[0],
                    "match_percent": match_check[1]
                },
                "missing": missing_tag_list,
                "detail": {extract_relative_tag(item[1]): item[0] for item in result}
            }

        except Exception as e:
            logging.error('Error at', exc_info=e)
        else:
            print(list2sentence(result))
            # return Response({"success": True, "data": result_data})



import logging
import time

import joblib

from .constants import tag_order, exactly_tags
# from .preprocess import process_csv_file
from .transform_to_dataset import transform_to_dataset
from .word2features import word2features

logger = logging.getLogger(__name__)


def pos_tag(model, sentence):
    sentence_features = [word2features(sentence, index) for index in range(len(sentence))]
    # print("model.predict([sentence_features])", model.predict([sentence_features]))
    return list(zip(sentence, model.predict([sentence_features])[0]))


def extract_relative_tag(tag):
    return tag.split('_')[0]


def extract_exactly_tag(tag):
    return tag.split('_')[1]


def sentence2list(sentence):
    return sentence.split(', ')


def list2sentence(list):
    result_list = []
    for item in list:
        relative_tag = extract_exactly_tag(item[1])
        # logger.info(item[0])
        name_list = item[0].split(" ")
        if len(name_list) > 1 and name_list[1].isdigit():
            address = item[0]
        else:
            address = "{} {}".format(exactly_tags[relative_tag], item[0])
        result_list.append(address)

    return ", ".join(result_list)


def crf_execute(model_path, sentence):
    # start = time.time()

    loaded_model = joblib.load(model_path)
    unsorted_result = pos_tag(loaded_model, sentence)
    sorted_result = sorted(unsorted_result, key=lambda x: tag_order[extract_relative_tag(x[1])], reverse=False)

    # end = time.time()
    # print('[{:.4f} s] Function crf_execute {} executed'.format((end - start), sentence))

    return sorted_result


# def crf_train(csv_path, model_path):
#     tagged_sentences, standard_addresses = process_csv_file(csv_path)
#     print("Tagged sentences: ", len(tagged_sentences))
#
#     # Split the dataset for training and testing
#     cutoff = int(1 * len(tagged_sentences))
#     training_sentences = tagged_sentences[:cutoff]
#     # test_sentences = tagged_sentences[cutoff:]
#
#     X_train, y_train = transform_to_dataset(training_sentences)
#     # X_test, y_test = transform_to_dataset(test_sentences)
#
#     from sklearn_crfsuite import CRF
#
#     model = CRF(
#         algorithm='lbfgs',
#         c1=0.1,
#         c2=0.1,
#         max_iterations=150,
#         # all_possible_transitions=True
#     )
#     model.fit(X_train, y_train)
#     joblib.dump(model, model_path)
#
#     return model_path, standard_addresses

# def full_crf_execute(train, model_path, sentence):
#     start = time.time()
#
#     if train:
#         tagged_sentences = process_csv_file(
#             "{}/{}/dvhc.csv".format(settings.BASE_DIR, os.environ.get("STANDARDIZER_FILE_PATH")))
#         print("Tagged sentences: ", len(tagged_sentences))
#
#         # Split the dataset for training and testing
#         cutoff = int(1 * len(tagged_sentences))
#         training_sentences = tagged_sentences[:cutoff]
#         test_sentences = tagged_sentences[cutoff:]
#
#         X_train, y_train = transform_to_dataset(training_sentences)
#         X_test, y_test = transform_to_dataset(test_sentences)
#
#         # print(len(X_train))
#         # print(len(X_test))
#         # print(X_train[0])
#         # print(y_train[0])
#
#         from sklearn_crfsuite import CRF
#
#         model = CRF(
#             algorithm='lbfgs',
#             c1=0.1,
#             c2=0.1,
#             max_iterations=150,
#             # all_possible_transitions=True
#         )
#         model.fit(X_train, y_train)
#         joblib.dump(model, model_path)
#         loaded_model = joblib.load(model_path)
#
#     else:
#         loaded_model = joblib.load(model_path)
#
#     unsorted_result = pos_tag(loaded_model, sentence)
#     sorted_result = sorted(unsorted_result, key=lambda x: tag_order[extract_relative_tag(x[1])], reverse=False)
#
#     end = time.time()
#     print('[{:.4f} s] Function crf_execute {} executed'.format((end - start), sentence))
#
#     return sorted_result



# import xlrd
# import csv
#
# def csv_from_excel(excel_path, sheet_name, csv_path):
#     wb = xlrd.open_workbook(excel_path)
#     sh = wb.sheet_by_name(sheet_name or "Sheet1")
#     your_csv_file = open(csv_path, 'w')
#     wr = csv.writer(your_csv_file, quoting=csv.QUOTE_ALL)
#
#     for rownum in range(sh.nrows)[1:]:
#         wr.writerow(sh.row_values(rownum))
#
#     your_csv_file.close()
#
# # runs the csv_from_excel function:
# # excel_path = "../../datas/Danh sách cấp tỉnh kèm theo quận huyện, phường xã ___23_04_2020.xls"
# # sheet_name = "Sheet1"
# # csv_path = "../../data/dvhc.csv"
# # csv_from_excel(excel_path, sheet_name, csv_path)


from .crf_execute import crf_execute
from .word2features import word2features
from .constants import exactly_tags, relative_tags, STANDARD_NAMES, tag_order
# from .preprocess import process_csv_file
from .transform_to_dataset import transform_to_dataset
# from .xlsx_to_csv import csv_from_excel


import csv
import re


def handle_numeric_name(raw_name, subfix_pattern):
    # name = raw_name.replace(subfix_pattern + " ", "")
    name = re.sub(subfix_pattern, "", raw_name, 1)
    if name.isdigit():
        return raw_name
    else:
        return name


def handle_province(raw_name):
    if re.compile("^(T|t)hành (P|p)hố ").match(raw_name):
        province_name = re.sub(r"^(T|t)hành (P|p)hố ", "", raw_name, 1)
        result_tuple = (province_name, "TTP_TP")

    elif re.compile("^(T|t)ỉnh ").match(raw_name):
        province_name = re.sub(r"^(T|t)ỉnh ", "", raw_name, 1)
        result_tuple = (province_name, "TTP_T")

    else:
        result_tuple = (raw_name, "TTP_T")

    return result_tuple


def handle_district(raw_name):
    if re.compile("^(Q|q)uận ").match(raw_name):
        result_tuple = (handle_numeric_name(raw_name, r"^(Q|q)uận "), "QH_Q")

    elif re.compile("^(H|h)uyện ").match(raw_name):
        district_name = re.sub(r"^(H|h)uyện ", "", raw_name, 1)
        result_tuple = (district_name, "QH_H")

    elif re.compile("^(T|t)hành (P|p)hố ").match(raw_name):
        district_name = re.sub(r"^(T|t)hành (P|p)hố ", "", raw_name, 1)
        result_tuple = (district_name, "QH_TP")

    elif re.compile("^(T|t)hị (X|x)ã ").match(raw_name):
        district_name = re.sub(r"^(T|t)hị (X|x)ã ", "", raw_name, 1)
        result_tuple = (district_name, "QH_TX")

    else:
        result_tuple = (raw_name, "QH_H")

    return result_tuple


def handle_commune(raw_name):
    if re.compile("^(X|x)ã ").match(raw_name):
        commune_name = re.sub(r"^(X|x)ã ", "", raw_name, 1)
        result_tuple = (commune_name, "PX_X")

    elif re.compile("^(P|p)hường ").match(raw_name):
        result_tuple = (handle_numeric_name(raw_name, r"^(P|p)hường "), "PX_P")

    elif re.compile("^(T|t)hị (T|t)rấn ").match(raw_name):
        commune_name = re.sub(r"^(T|t)hị (T|t)rấn ", "", raw_name, 1)
        result_tuple = (commune_name, "PX_TT")

    else:
        result_tuple = (raw_name, "PX_X")

    return result_tuple


import random


def process_csv_file(csv_path):
    result = []
    standard_addresses = []
    with open(csv_path) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        tmp_province_code = 0
        tmp_district_code = 0

        for index, row in enumerate(csv_reader):
            row_data = []
            # if row[1] != tmp_province_code:
            #     tmp_province_code = row[1]
            #     result.append([handle_province(row[0])])
            #
            # if row[3] != tmp_district_code:
            #     tmp_district_code = row[3]
            #     result.append([handle_district(row[2])])

            province = handle_province(row[0])
            district = handle_district(row[2])
            commune = handle_commune(row[4])

            # with open('../../datas/standard.txt', 'a') as name_file:
            #     name_file.write('\"{}, {}, {}\",\n'.format(commune[0], district[0], province[0]))

            row_data.append(province) if province[0] != "" else None
            row_data.append(district) if district[0] != "" else None
            row_data.append(commune) if commune[0] != "" else None
            random.shuffle(row_data)

            standard_addresses.append("{}, {}, {}".format(commune[0], district[0], province[0]))
            result.append(row_data)

    # print(result[0])
    # print(len(result))
    return result, standard_addresses

# process_csv_file("/home/vantrong291/workspaces/thesis/weaddress/backend/standardizer/datas/dvhc.csv")



import time

from .constants import ACRONYM_DICT


def acronym_execute(raw_address):
    # start = time.time()

    address = raw_address
    for item in ACRONYM_DICT:
        if item in raw_address:
            address = address.replace(item, ACRONYM_DICT[item])

    # end = time.time()
    # print('[{:.4f} s] Function acronym_execute {} executed'.format((end - start), raw_address))
    return address



def word2features(sentence, index):
    # print(index)
    return {
        # 'character_length': len(sentence[index]),
        # 'word_length': sentence[index].count(" ") + 1,
        'word': sentence[index],
        'is_first': index == 0,
        'is_last': index == len(sentence) - 1,
        # 'is_capitalized': sentence[index][0].upper() == sentence[index][0],
        # 'is_all_upper': sentence[index].upper() == sentence[index],
        # 'is_all_lower': sentence[index].lower() == sentence[index],
        'prefix-1': sentence[index][0],
        'prefix-2': sentence[index][:2],
        'prefix-3': sentence[index][:3],
        'prefix-4': sentence[index][:4],
        'prefix-5': sentence[index][:5],
        'suffix-1': sentence[index][-1],
        'suffix-2': sentence[index][-2:],
        'suffix-3': sentence[index][-3:],
        'suffix-4': sentence[index][-4:],
        'suffix-5': sentence[index][-5:],
        'prev_word': '' if index == 0 else sentence[index - 1],
        'next_word': '' if index == len(sentence) - 1 else sentence[index + 1],
        'has_hyphen': '-' in sentence[index],
        'is_numeric': sentence[index].isdigit(),
        'capitals_inside': sentence[index][1:].lower() != sentence[index][1:]
    }



# from crf_layer.crf_execute import crf_execute, list2sentence
# from fuzzywuzzy import fuzz, process
# from spell_correction_layer.generate_error_dict import generate_error_dict
# from spell_correction_layer.match_word import match_word
# from utils import sentence2list
#
# from crf_layer.constants import STANDARD_NAMES
#
#
# def combined_process(address, crf_train = False,):
#     address_list = ["Hải Hậuu, Hari Minh, Namdinh", "Khắc Niệm, Bắc Ninh, Bắc Ninh", "Hải Hậu, Nam Định, Cồn",
#                     "Ba Đình, Hà Nội, Phúc Xá", "Hà Nội", "Mai Dịch, Hà Nội, Cầu Giấy",
#                     "9, Phước Long B, Hồ Chí Minh", "Minh Khai, Hà Nội",
#                     "Nam Đàn, Nghệ An", "Đầm Dơi, Cà Mau, Taj An Khuwowng Nam", "Ninh Kiềuu, Cần Thơ"]
#     for index, address in enumerate(address_list):
#         address2list = sentence2list(address)
#         error_dict = generate_error_dict()
#         corrected_address2list = []
#         for name in address2list:
#             corrected_name = match_word(error_dict, name)
#             corrected_address2list.append(corrected_name)
#
#         crf_result = crf_execute(crf_train, "../models/finalized_model.sav", corrected_address2list)
#         name_only_crf_result = ", ".join([item[0] for item in crf_result])
#         # print(name_only_crf_result)
#         match_check = process.extractOne(name_only_crf_result, STANDARD_NAMES, scorer=fuzz.ratio)
#
#         print("{}. {} -> {}, match {}%".format(index + 1, address, list2sentence(crf_result), match_check[1]))
#
# combined_process("")


from nltk.tag.util import untag
from .word2features import word2features


def transform_to_dataset(tagged_sentences):
    X, y = [], []

    for tagged in tagged_sentences:
        X.append([word2features(untag(tagged), index) for index in range(len(tagged))])
        y.append([tag for _, tag in tagged])

    # print("y")
    # print(y)

    return X, y



ACRONYM_DICT = {
    "HN": "Hà Nội",
    "TPHCM": "Hồ Chí Minh",
    "TP.HCM": "Hồ Chí Minh",
    "HP": "Hải Phòng",
    "ĐN": "Đà Nắng",
    "DN": "Đà Nắng",
    "CT": "Cần Thơ",
    "NĐ": "Nam Định",
    "ND": "Nam Định"
}


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='vnaddress',
    version='0.9',
    scripts=['vnas'],
    author="vantrong291",
    author_email="vantrong291@gmail.com",
    description="A package for parsing Vietnamese address",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vantrong291/vn_address_standardizer",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'nltk', 'joblib', 'sklearn_crfsuite', 'fuzzywuzzy', 'python-Levenshtein'
    ],
    include_package_data=True,
)



# import csv
# import os
# import re
#
# from django.conf import settingssettings
#
#
# def handle_numeric_name(raw_name, subfix_pattern):
#     # name = raw_name.replace(subfix_pattern + " ", "")
#     name = re.sub(subfix_pattern, "", raw_name, 1)
#     if name.isdigit():
#         return raw_name
#     else:
#         return name
#
#
# def handle_province(raw_name):
#     if re.compile("^(T|t)hành (P|p)hố ").match(raw_name):
#         province_name = re.sub(r"^(T|t)hành (P|p)hố ", "", raw_name, 1)
#
#     elif re.compile("^(T|t)ỉnh ").match(raw_name):
#         province_name = re.sub(r"^(T|t)ỉnh ", "", raw_name, 1)
#
#     else:
#         province_name = raw_name
#
#     return province_name
#
#
# def handle_district(raw_name):
#     if re.compile("^(Q|q)uận ").match(raw_name):
#         district_name = handle_numeric_name(raw_name, r"^(Q|q)uận ")
#
#     elif re.compile("^(H|h)uyện ").match(raw_name):
#         district_name = re.sub(r"^(H|h)uyện ", "", raw_name, 1)
#         # result_tuple = (district_name, "QH_H")
#
#     elif re.compile("^(T|t)hành (P|p)hố ").match(raw_name):
#         district_name = re.sub(r"^(T|t)hành (P|p)hố ", "", raw_name, 1)
#
#     elif re.compile("^(T|t)hị (X|x)ã ").match(raw_name):
#         district_name = re.sub(r"^(T|t)hị (X|x)ã ", "", raw_name, 1)
#
#     else:
#         district_name = raw_name
#
#     return district_name
#
#
# def handle_commune(raw_name):
#     if re.compile("^(X|x)ã ").match(raw_name):
#         commune_name = re.sub(r"^(X|x)ã ", "", raw_name, 1)
#
#     elif re.compile("^(P|p)hường ").match(raw_name):
#         commune_name = handle_numeric_name(raw_name, r"^(P|p)hường ")
#
#     elif re.compile("^(T|t)hị (T|t)rấn ").match(raw_name):
#         commune_name = re.sub(r"^(T|t)hị (T|t)rấn ", "", raw_name, 1)
#
#     else:
#         commune_name = raw_name
#
#     return commune_name
#
#
# def process_csv_file(csv_path):
#     print(csv_path)
#     result = []
#     with open(csv_path) as csv_file:
#         csv_reader = csv.reader(csv_file, delimiter=',')
#         for index, row in enumerate(csv_reader):
#             province = handle_province(row[0])
#             district = handle_district(row[2])
#             commune = handle_commune(row[4])
#
#             result.append(province) if province != "" and province not in result else None
#             result.append(district) if district != "" and district not in result else None
#             result.append(commune) if commune != "" and commune not in result else None
#
#     with open('{}/{}/dvhc.txt'.format(settings.BASE_DIR, os.environ.get("STANDARDIZER_FILE_PATH")), 'w') as dvhc:
#         for word in result:
#             dvhc.write('{}\n'.format(word))
#
#     with open('{}/{}/names.txt'.format(settings.BASE_DIR, os.environ.get("STANDARDIZER_FILE_PATH")), 'w') as name_file:
#         names = []
#         for word in result:
#             names.append("{}".format(word))
#         name_file.write('{}'.format(names))
#
#     print(len(result))
#     return result
#
